# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager

import pytest

from pallas.exceptions import AthenaQueryError
from pallas.interruptions import AthenaKillOnInterruptWrapper
from pallas.testing import AthenaFake, QueryFake


class InterruptQueryFake(QueryFake):
    _interrupted: bool = False

    def join(self) -> None:
        if not self._interrupted:
            self._interrupted = True
            raise KeyboardInterrupt
        super().join()


class InterruptAthenaFake(AthenaFake):

    state = "CANCELLED"
    query_cls = InterruptQueryFake


@pytest.fixture(name="athena")
def athena_fixture():
    fake = InterruptAthenaFake()
    return AthenaKillOnInterruptWrapper(fake)


@contextmanager
def no_keyboard_interrupt():
    """
    Convert KeyboardInterrupt to test failures.

    When KeyboardInterrupt is not caught by AthenaKillOnInterruptWrapper,
    do not want to terminate the pytest process.
    """
    try:
        yield
    except KeyboardInterrupt:
        raise AssertionError("Unexpected KeyboardInterrupt")


class TestAthenaKillOnInterruptWrapper:
    def test_execute(self, athena):
        with no_keyboard_interrupt():
            with pytest.raises(AthenaQueryError):
                athena.execute("SELECT 1")
            assert athena.wrapped.request_log == [
                "StartQueryExecution",
                "StopQueryExecution",
                "GetQueryExecution",
            ]

    def test_get_results(self, athena):
        with no_keyboard_interrupt():
            query = athena.submit("SELECT 1")
            athena.wrapped.request_log.clear()
            with pytest.raises(AthenaQueryError):
                query.get_results()
            assert athena.wrapped.request_log == [
                "StopQueryExecution",
                "GetQueryExecution",
            ]

    def test_join(self, athena):
        with no_keyboard_interrupt():
            query = athena.submit("SELECT 1")
            athena.wrapped.request_log.clear()
            with pytest.raises(AthenaQueryError):
                query.join()
            assert athena.wrapped.request_log == [
                "StopQueryExecution",
                "GetQueryExecution",
            ]



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Decorators for caching Athena queries.
"""

from __future__ import annotations

import hashlib
import logging
import re
from typing import Optional
from urllib.parse import urlencode

from pallas.base import Athena, AthenaWrapper, Query, QueryWrapper
from pallas.results import QueryResults
from pallas.storage import NotFoundError, Storage

logger = logging.getLogger("pallas")


_comment_1 = r"--[^\n]*\n"
_comment_2 = r"/\*([^*]|\*(?!/))*\*/"

SELECT_RE = re.compile(
    rf"(\s+|{_comment_1}|{_comment_2})*(SELECT|WITH)\b", re.IGNORECASE
)


def is_cacheable(sql: str) -> bool:
    """
    Return whether an SQL statement can be cached.

    Only SELECT statements are considered cacheable.
    """
    return SELECT_RE.match(sql) is not None


class AthenaCachingWrapper(AthenaWrapper):
    """
    Athena wrapper that caches query IDs and optionally results.

    Athena always stores results in S3, so it is possible
    to retrieve past results without manually copying data.

    The caching wrapper can work two modes,
    depending on *cache_results* setting:

    1) Remote mode - cache query execution IDs.
    2) Local mode - Cache query execution IDs and query results.

    I the remote mode, cache is used to recover query execution IDs
    that are necessary for downloading results of past queries.

    In the local mode, results are stored to cache too.
    If the cache storage is local then it is possible to retrieve
    cached results without internet connection.
    """

    _storage: Storage

    def __init__(
        self, wrapped: Athena, *, storage: Storage, cache_results: bool = True
    ) -> None:
        super().__init__(wrapped)
        self._storage = storage
        self._cache_results = cache_results

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__}: {self.wrapped!r} cached at {self.storage.uri!r}>"
        )

    @property
    def storage(self) -> Storage:
        return self._storage

    @property
    def cache_results(self) -> bool:
        return self._cache_results

    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        sql_cacheable = is_cacheable(sql)
        if sql_cacheable and not ignore_cache:
            execution_id = self._load_execution_id(sql)
            if execution_id is not None:
                return self.get_query(execution_id)
        query = super().submit(sql, ignore_cache=ignore_cache)
        if sql_cacheable:
            self._save_execution_id(sql, query.execution_id)
            query = self._wrap_query(query)
        return query

    def get_query(self, execution_id: str) -> Query:
        query = super().get_query(execution_id)
        return self._wrap_query(query)

    def _wrap_query(self, query: Query) -> Query:
        if self._cache_results:
            query = QueryCachingWrapper(query, self._storage)
        return query

    def _load_execution_id(self, sql: str) -> Optional[str]:
        key = self._get_cache_key(sql)
        try:
            execution_id = self._storage.get(key)
        except NotFoundError:
            return None
        logger.info(
            f"Query execution loaded from cache {self._storage}{key}:"
            f" QueryExecutionId={execution_id!r}"
        )
        return execution_id

    def _save_execution_id(self, sql: str, execution_id: str) -> None:
        key = self._get_cache_key(sql)
        self._storage.set(key, execution_id)
        logger.info(
            f"Query execution saved to cache {self._storage}{key}:"
            f" QueryExecutionId={execution_id!r}"
        )

    def _get_cache_key(self, sql: str) -> str:
        parts = {}
        if self.database is not None:
            parts["database"] = self.database
        parts["sql"] = sql
        encoded = urlencode(parts).encode("utf-8")
        hash = hashlib.sha256(encoded).hexdigest()
        return f"query-{hash}"


class QueryCachingWrapper(QueryWrapper):
    """
    Query wrapper that caches query results.
    """

    _storage: Storage

    def __init__(self, wrapped: Query, storage: Storage) -> None:
        super().__init__(wrapped)
        self._storage = storage

    def get_results(self) -> QueryResults:
        results = self._load_results()
        if results is not None:
            return results
        results = super().get_results()
        # When an old query is accessed using its execution ID,
        # we have to retrieve its SQL to determine whether it can be cached.
        # Network should not be used until a cache is checked.
        # Getting the SQL here should not produce an additional request.
        if is_cacheable(self.get_info().sql):
            self._save_results(results)
        return results

    def join(self) -> None:
        if self._has_results():
            # If we have results then we can assume that query has finished.
            return
        super().join()

    def _has_results(self) -> bool:
        return self._storage.has(self._get_cache_key())

    def _load_results(self) -> Optional[QueryResults]:
        key = self._get_cache_key()
        try:
            with self._storage.reader(key) as stream:
                results = QueryResults.load(stream)
        except NotFoundError:
            return None
        logger.info(
            f"Query results loaded from cache {self._storage}{key}:"
            f" QueryExecutionId={self.execution_id!r}: {len(results)} rows"
        )
        return results

    def _save_results(self, results: QueryResults) -> None:
        key = self._get_cache_key()
        with self._storage.writer(key) as stream:
            results.save(stream)
        logger.info(
            f"Query results saved to cache {self._storage}{key}:"
            f" QueryExecutionId={self.execution_id!r}: {len(results)} rows"
        )

    def _get_cache_key(self) -> str:
        return f"results-{self.execution_id}"



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
from decimal import Decimal

import numpy as np
import pytest

from pallas.base import Athena


@pytest.mark.parametrize(
    "value,quoted",
    [
        (None, "NULL"),
        ("", "''"),
        ("hello", "'hello'"),
        ("'", "''''"),
        (42, "42"),
        (np.int64(42), "42"),
        (3.14, "3.14"),
        (float("nan"), "nan()"),
        (float("inf"), "infinity()"),
        (float("-inf"), "-infinity()"),
        (np.float64(3.14), "3.14"),
        (np.float64("nan"), "nan()"),
        (np.float64("inf"), "infinity()"),
        (np.float64("-inf"), "-infinity()"),
        (np.nan, "nan()"),
        (np.inf, "infinity()"),
        (-np.inf, "-infinity()"),
        (Decimal("0.1"), "DECIMAL '0.1'"),
        (dt.date(2001, 8, 22), "DATE '2001-08-22'"),
        (
            dt.datetime(2001, 8, 22, 3, 4, 5, 321000),
            "TIMESTAMP '2001-08-22 03:04:05.321'",
        ),
        (b"\x00\xff\x00", "from_base64('AP8A')"),
    ],
)
def test_quote_value(value, quoted):
    assert Athena.quote(value) == quoted


@pytest.mark.parametrize("value", [object(), (), [], {}])
def test_quote_invalid(value):
    with pytest.raises(TypeError):
        Athena.quote(value)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Facade that can setup an Athena clients from configuration.

Full-features Athena client is made a layered composition of objects.
The core is an :class:`.AthenaProxy` instance that issues requests to AWS.
It is wrapped by decorators for extra functionality like caching.

Function defined in this module can assembly the whole object structure.
"""

from __future__ import annotations

import os
from typing import Mapping, Optional

from pallas.base import Athena
from pallas.caching import AthenaCachingWrapper
from pallas.interruptions import AthenaKillOnInterruptWrapper
from pallas.normalization import AthenaNormalizationWrapper
from pallas.proxies import AthenaProxy
from pallas.storage import storage_from_uri


def setup(
    *,
    database: Optional[str] = None,
    workgroup: Optional[str] = None,
    output_location: Optional[str] = None,
    region: Optional[str] = None,
    cache_remote: Optional[str] = None,
    cache_local: Optional[str] = None,
    normalize: bool = True,
    kill_on_interrupt: bool = True,
) -> Athena:
    """
    Setup :class:`.Athena` instance.

    Initializes :class:`.AthenaProxy` and decorates it by caching wrappers.

    :param environ: Mapping to use instead of ``os.environ``.
        Set to empty dict to ignore environment variables.
    :param environ_prefix: Prefix of environment variables.
    :param database: a name of Athena database.
        If omitted, database should be specified in SQL.
    :param workgroup: a name of Athena workgroup.
        If omitted, default workgroup will be used.
    :param output_location: an output location at S3 for query results.
        Optional if a default location is specified for the *workgroup*.
    :param region: an AWS region.
        By default, region from AWS config is used.
    :param cache_remote: an URI of a remote cache.
        Query execution IDs without results are stored to the remote cache.
    :param cache_local: an URI of a local cache.
        Both results and query execution IDs are stored to the local cache.
    :param normalize: whether to normalize SQL
        Normalizes whitespace to improve caching.
    :param kill_on_interrupt: whether to kill queries on KeyboardInterrupt
        Kills query when interrupted during waiting.
    :return: an Athena instance
        A :class:`.AthenaProxy` instance wrapped necessary in decorators.
    """
    athena: Athena = AthenaProxy(
        database=database,
        workgroup=workgroup,
        output_location=output_location,
        region=region,
    )
    if cache_remote is not None:
        storage = storage_from_uri(cache_remote)
        athena = AthenaCachingWrapper(athena, storage=storage, cache_results=False)
    if cache_local is not None:
        storage = storage_from_uri(cache_local)
        athena = AthenaCachingWrapper(athena, storage=storage, cache_results=True)
    if normalize:
        athena = AthenaNormalizationWrapper(athena)
    if kill_on_interrupt:
        athena = AthenaKillOnInterruptWrapper(athena)
    return athena


def environ_setup(
    environ: Optional[Mapping[str, str]] = None, *, prefix: str = "PALLAS"
) -> Athena:
    """
    Setup :class:`.Athena` instance from environment variables.

    Reads the following environment variables: ::

        export PALLAS_DATABASE=
        export PALLAS_WORKGROUP=
        export PALLAS_OUTPUT_LOCATION=
        export PALLAS_REGION=
        export PALLAS_NORMALIZE=true
        export PALLAS_KILL_ON_INTERRUPT=true
        export PALLAS_CACHE_REMOTE=$PALLAS_OUTPUT_LOCATION
        export PALLAS_CACHE_LOCAL=~/Notebooks/.cache/

    :param environ: A mapping object representing the string environment.
        Defaults to ``os.environ``.
    :param prefix: Prefix of environment variables.
    :return: an Athena instance
        A :class:`.AthenaProxy` instance wrapped necessary in decorators.
    """
    if environ is None:
        environ = os.environ
    config = _EnvironConfig(environ, prefix)
    return setup(
        database=config.get_str("DATABASE"),
        workgroup=config.get_str("WORKGROUP"),
        output_location=config.get_str("OUTPUT_LOCATION"),
        region=config.get_str("REGION"),
        cache_remote=config.get_str("CACHE_REMOTE"),
        cache_local=config.get_str("CACHE_LOCAL"),
        normalize=config.get_bool("NORMALIZE", True),
        kill_on_interrupt=config.get_bool("KILL_ON_INTERRUPT", True),
    )


class _EnvironConfig:
    def __init__(self, environ: Mapping[str, str], prefix: str) -> None:
        self._environ = environ
        self._prefix = prefix

    def get_str(self, key: str, default: Optional[str] = None) -> Optional[str]:
        v = self._get(key)
        if not v:
            return default
        return v

    def get_bool(self, key: str, default: bool = False) -> bool:
        v = self._get(key)
        if not v:
            return default
        v = v.lower()
        if v in ("1", "true", "on", "yes"):
            return True
        if v in ("0", "false", "off", "no"):
            return False
        raise ValueError(f"{self._prefix}_{key}: invalid boolean value: {v}")

    def _get(self, key: str) -> str:
        v = self._environ.get(f"{self._prefix}_{key}", "")
        if not isinstance(v, str):
            # Avoid unexpected behaviour when somebody passes something
            # like {"PALLAS_KILL_ON_INTERRUPT": False} to environ.
            raise TypeError("Environ values must be a string.")
        return v



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Normalization of Athena queries
"""

from __future__ import annotations

import textwrap

from pallas.base import AthenaWrapper, Query


def normalize_sql(sql: str) -> str:
    lines = sql.splitlines()
    joined = "\n".join(line.rstrip() for line in lines)
    return textwrap.dedent(joined).strip()


class AthenaNormalizationWrapper(AthenaWrapper):
    """
    Athena wrapper that normalizes executed queries.

    Query normalization can improve caching.

    Following normalization operations are done:
    - Common indentation is removed.
    - Heading and trailing new lines are removed.
    - Trailing whitespace is removed from end of lines.
    - Line endings are normalized to LF
    """

    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        normalized = normalize_sql(sql)
        return super().submit(normalized, ignore_cache=ignore_cache)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Proxies to AWS Athena APIs.

Classes implemented in this module are the core of the Pallas library.
They issue actual requests using boto3.

Other clients are implemented as decorators wrapping the core
and offering same interface.
"""

from __future__ import annotations

import logging
import time
from typing import Any, Dict, Mapping, Optional, Sequence

import boto3

from pallas.base import Athena, Query
from pallas.csv import read_csv
from pallas.info import QueryInfo
from pallas.results import QueryResults
from pallas.storage.s3 import s3_parse_uri, s3_wrap_body
from pallas.utils import Fibonacci, truncate_str

logger = logging.getLogger("pallas")


ColumnNames = Sequence[str]
ColumnTypes = Sequence[str]
Row = Sequence[Optional[str]]


class AthenaProxy(Athena):
    """
    Proxy to AWS Athena.

    This is the core implementation of the :class:`.Athena` interface.
    It executes queries via AWS APIs using boto3 library.

    It can be decorated by wrappers to provide additional functionality.

    :param database: a name of Athena database.
        If omitted, database should be specified in SQL.
    :param workgroup: a name of Athena workgroup.
        If omitted, default workgroup will be used.
    :param output_location: an output location at S3 for query results.
        Optional if a default location is specified for the *workgroup*.
    :param region: an AWS region.
        By default, a region from AWS config is used.
    :param athena_client: a boto3 client to use.
        By default, a new client is constructed.
    :param s3_client: a boto3 client to use.
        By default, a new client is constructed.
    """

    _athena_client: Any  # boto3 Athena client
    _s3_client: Any  # boto3 S3 client
    _database: Optional[str]
    _workgroup: Optional[str]
    _output_location: Optional[str]

    def __init__(
        self,
        *,
        database: Optional[str] = None,
        workgroup: Optional[str] = None,
        output_location: Optional[str] = None,
        region: Optional[str] = None,
        athena_client: Optional[Any] = None,
        s3_client: Optional[Any] = None,
    ) -> None:
        if athena_client is None:
            athena_client = boto3.client("athena", region_name=region)
        if s3_client is None:
            s3_client = boto3.client("s3")
        self._athena_client = athena_client
        self._s3_client = s3_client
        self._database = database
        self._workgroup = workgroup
        self._output_location = output_location

    def __repr__(self) -> str:
        parts = []
        if self.database is not None:
            parts.append(f"database={self.database!r}")
        if self.output_location is not None:
            parts.append(f"output_location={self.output_location!r}")
        return f"<{type(self).__name__}: {', '.join(parts)}>"

    @property
    def database(self) -> Optional[str]:
        return self._database

    @property
    def workgroup(self) -> Optional[str]:
        return self._workgroup

    @property
    def output_location(self) -> Optional[str]:
        return self._output_location

    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        params: Dict[str, Any] = dict(QueryString=sql)
        if self._database is not None:
            params.update(QueryExecutionContext={"Database": self._database})
        if self._workgroup is not None:
            params.update(WorkGroup=self._workgroup)
        if self._output_location is not None:
            params.update(ResultConfiguration={"OutputLocation": self._output_location})
        logger.info(f"Athena StartQueryExecution: QueryString={truncate_str(sql)!r}")
        response = self._athena_client.start_query_execution(**params)
        query = self.get_query(response["QueryExecutionId"])
        logger.info(f"Athena QueryExecutionId={query.execution_id!r} started.")
        return query

    def get_query(self, execution_id: str) -> Query:
        return QueryProxy(
            execution_id, athena_client=self._athena_client, s3_client=self._s3_client
        )


class QueryProxy(Query):
    """
    Proxy to an Athena query execution.

    This is the core implementation of the :class:`.Query` interface.
    It can monitor and control the query execution via AWS APIs
    using boto3 library.

    It can be decorated by wrappers to provide additional functionality.
    """

    _athena_client: Any  # boto3 Athena client
    _s3_client: Any  # boto3 S3 client
    _execution_id: str
    _finished_info: Optional[QueryInfo] = None

    def __init__(
        self, execution_id: str, *, athena_client: Any, s3_client: Any
    ) -> None:
        self._athena_client = athena_client
        self._s3_client = s3_client
        self._execution_id = execution_id

    @property
    def execution_id(self) -> str:
        return self._execution_id

    def get_info(self) -> QueryInfo:
        # Query info is cached if the query finished so it cannot change.
        if self._finished_info is not None:
            return self._finished_info
        logger.info(f"Athena GetQueryExecution: QueryExecutionId={self.execution_id!r}")
        response = self._athena_client.get_query_execution(
            QueryExecutionId=self.execution_id
        )
        info = QueryInfo(response["QueryExecution"])
        logger.info(f"Athena QueryExecution: {info}")
        if info.finished:
            self._finished_info = info
        return info

    def get_results(self) -> QueryResults:
        self.join()
        params = dict(QueryExecutionId=self.execution_id)
        logger.info(f"Athena GetQueryResults: QueryExecutionId={self.execution_id!r}")
        response = self._athena_client.get_query_results(**params)
        column_names = self._read_column_names(response)
        column_types = self._read_column_types(response)
        if response.get("NextToken"):
            logger.info("Athena ResultSet paginated. Will download from S3.")
            data = self._download_data()
        else:
            data = self._read_data(response)
            logger.info(
                f"Athena ResultSet complete: {len(data)} rows (including header)"
            )
        fixed_data = _fix_data(column_names, data)
        return QueryResults(column_names, column_types, fixed_data)

    def kill(self) -> None:
        logger.info(
            f"Athena StopQueryExecution: QueryExecutionId={self.execution_id!r}"
        )
        self._athena_client.stop_query_execution(QueryExecutionId=self.execution_id)

    def join(self) -> None:
        for delay in Fibonacci(max_value=60):
            info = self.get_info()
            if info.finished:
                info.check()
                break
            time.sleep(delay)

    def _read_column_names(self, response: Mapping[str, Any]) -> ColumnNames:
        column_info = response["ResultSet"]["ResultSetMetadata"]["ColumnInfo"]
        return tuple(column["Name"] for column in column_info)

    def _read_column_types(self, response: Mapping[str, Any]) -> ColumnTypes:
        column_info = response["ResultSet"]["ResultSetMetadata"]["ColumnInfo"]
        return tuple(column["Type"] for column in column_info)

    def _read_data(self, response: Mapping[str, Any]) -> Sequence[Row]:
        rows = response["ResultSet"]["Rows"]
        return [tuple(item.get("VarCharValue") for item in row["Data"]) for row in rows]

    def _download_data(self) -> Sequence[Row]:
        output_location = self.get_info().output_location
        bucket, key = s3_parse_uri(output_location)
        params = dict(Bucket=bucket, Key=key)
        logger.info(f"S3 GetObject:" f" Bucket={bucket!r} Key={key!r}")
        response = self._s3_client.get_object(**params)
        with s3_wrap_body(response["Body"]) as stream:
            data = list(read_csv(stream))
        logger.info(f"S3 Body downloaded: {len(data)} rows (including header)")
        return data


def _fix_data(column_names: ColumnNames, data: Sequence[Row]) -> Sequence[Row]:
    """
    Fix malformed data returned from Athena.

    Queries executed by Presto (typically queries with SELECT)
    repeat column names in the first row of data,
    so we have to remove them.

    Queries by Hive (typically queries with DESCRIBE)
    do not repeat column names, but all columns are combined to one.

    Try to fix both of the above problems here.
    """
    if data and data[0] == column_names:
        # DQL, SELECT statements executed by Presto
        data = data[1:]
    elif all(len(row) == 1 for row in data) and len(column_names) > 1:
        # DCL, DESCRIBE statements executed by Hive
        values = (row[0] for row in data if row[0] is not None)
        data = [v.split("\t", maxsplit=len(column_names) - 1) for v in values]
    return data



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pallas.normalization import AthenaNormalizationWrapper, normalize_sql
from pallas.testing import AthenaFake

NORMALIZED_SQL = """\
SELECT
    c1, c2
FROM
    t\
"""


class TestNormalizeSQL:
    def test_dedent(self):
        sql = normalize_sql(
            """
            SELECT
                c1, c2
            FROM
                t
        """
        )
        assert sql == NORMALIZED_SQL

    def test_trailing_whitespace(self):
        sql = normalize_sql(
            """\N{space}
            SELECT\N{space}\N{space}\N{space}
                c1, c2\N{space}
            FROM\N{space}
                t\N{space}
        """
        )
        assert sql == NORMALIZED_SQL

    def test_empty_lines(self):
        # Leading and trailing new lines are removed.
        # Other new lines are normalized to LF
        sql = normalize_sql(
            """\n\n\n
            SELECT\n\N{space}\N{space}\N{space}\n\r\n
                c1, c2
            FROM
                t\n\n\n
        """
        )
        assert sql == NORMALIZED_SQL.replace("SELECT", "SELECT\n\n\n")


@pytest.fixture(name="athena")
def athena_fixture():
    fake = AthenaFake()
    fake.database = "test_database"
    return AthenaNormalizationWrapper(fake)


class TestAthenaNormalizationWrapper:
    def test_repr(self, athena):
        assert repr(athena) == "<AthenaNormalizationWrapper: <AthenaFake>>"

    def test_query_repr(self, athena):
        query = athena.submit("SELECT 1")
        assert repr(query) == "<QueryFake: execution_id='query-1'>"

    def test_database(self, athena):
        assert athena.database == "test_database"

    def test_submit(self, athena):
        query = athena.submit(
            """
            SELECT
                c1, c2
            FROM
                t
        """
        )
        assert query.get_info().sql == NORMALIZED_SQL

    def test_get_query(self, athena):
        athena.submit(
            """
            SELECT
                c1, c2
            FROM
                t
        """
        )
        query = athena.get_query("query-1")
        assert query.get_info().sql == NORMALIZED_SQL



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Interruption handling of Athena queries
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import Iterator

from pallas.base import AthenaWrapper, Query, QueryWrapper
from pallas.results import QueryResults


@contextmanager
def _kill_on_interrupt(query: Query) -> Iterator[None]:
    try:
        yield
    except KeyboardInterrupt:
        query.kill()
        query.join()  # Wait until killed and raise


class QueryKillOnInterruptWrapper(QueryWrapper):
    """
    Query wrapper that kills queries on the KeyboardInterrupt exception.
    """

    def get_results(self) -> QueryResults:
        with _kill_on_interrupt(self.wrapped):
            return self.wrapped.get_results()

    def join(self) -> None:
        with _kill_on_interrupt(self.wrapped):
            return self.wrapped.join()


class AthenaKillOnInterruptWrapper(AthenaWrapper):
    """
    Athena wrapper that kills queries on the KeyboardInterrupt exception.
    """

    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        query = super().submit(sql, ignore_cache=ignore_cache)
        return self._wrap_query(query)

    def get_query(self, execution_id: str) -> Query:
        query = super().get_query(execution_id)
        return self._wrap_query(query)

    def _wrap_query(self, query: Query) -> Query:
        return QueryKillOnInterruptWrapper(query)



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Exceptions raised when querying Athena.
"""

from __future__ import annotations

from typing import Optional


class AthenaQueryError(Exception):
    """Athena query failed."""

    def __init__(self, state: str, state_reason: Optional[str]):
        self.state = state
        self.state_reason = state_reason

    def __str__(self) -> str:
        if self.state_reason is not None:
            return f"Athena query {self.state.lower()}: {self.state_reason}"
        return f"Athena query {self.state.lower()}"



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SQL helpers.

Implements quoting compatible with Athena.
"""

import base64
import datetime as dt
import math
import numbers
from decimal import Decimal
from typing import Union

SQL_SCALAR = Union[None, str, float, numbers.Real, Decimal, bytes, dt.date]


def _quote_str(value: str) -> str:
    escaped = str(value).replace("'", "''")
    return f"'{escaped}'"


def _quote_integral(value: Union[int, numbers.Integral]) -> str:
    if not isinstance(value, int):
        value = int(value)
    return str(value)


def _quote_real(value: Union[float, numbers.Real]) -> str:
    if not isinstance(value, float):
        value = float(value)
    if math.isnan(value):
        return "nan()"
    if math.isinf(value):
        return "infinity()" if value > 0 else "-infinity()"
    return str(value)


def _quote_decimal(value: Decimal) -> str:
    return f"DECIMAL {_quote_str(str(value))}"


def _quote_bytes(value: bytes) -> str:
    encoded = base64.b64encode(value).decode("ascii")
    return f"from_base64({_quote_str(encoded)})"


def _quote_datetime(value: dt.datetime) -> str:
    encoded = value.isoformat(sep=" ", timespec="milliseconds")
    return f"TIMESTAMP {_quote_str(encoded)}"


def _quote_date(value: dt.date) -> str:
    encoded = value.isoformat()
    return f"DATE {_quote_str(encoded)}"


def quote(value: SQL_SCALAR) -> str:
    """
    Quote scalar method to an SQL expression.

    :param value: Python value
    :return: SQL expression
    """
    if value is None:
        return "NULL"
    if isinstance(value, str):
        return _quote_str(value)
    if isinstance(value, int) or isinstance(value, numbers.Integral):
        return _quote_integral(value)
    if isinstance(value, float) or isinstance(value, numbers.Real):
        return _quote_real(value)
    if isinstance(value, Decimal):
        return _quote_decimal(value)
    if isinstance(value, dt.datetime):
        return _quote_datetime(value)
    if isinstance(value, dt.date):
        return _quote_date(value)
    if isinstance(value, bytes):
        return _quote_bytes(value)
    raise TypeError(f"Cannot quote {type(value)}.")



# Copyright 2020 Akamai Technologies, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Interfaces and base classes for querying AWS Athena.

Pallas exposes to main interfaces:
 - :class:`.Athena` for submitting queries.
 - :class:`.Query` for non-blocking monitoring of query status.

Classes :class:`.AthenaProxy` and :class`.QueryProxy`
are the core implementations of the interfaces.
They can be optionally decorated by wrappers that provide
extra functionality (like caching) preserving the same interface.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Optional

from pallas.info import QueryInfo
from pallas.results import QueryResults
from pallas.sql import quote


class Athena(metaclass=ABCMeta):
    """
    Athena interface

    Provides methods to execute SQL queries in AWS Athena.

    This is an abstract base class.
    The :class:`.AthenaProxy` subclass submits actual queries to AWS API.
    Other subclasses are implemented as decorators adding extra
    functionally (for example caching).
    """

    def __repr__(self) -> str:
        return f"<{type(self).__name__}>"

    quote = staticmethod(quote)

    @property
    @abstractmethod
    def database(self) -> Optional[str]:
        """
        Name of Athena database that will be queries.

        Individual queries can override this in SQL.
        """

    @property
    @abstractmethod
    def workgroup(self) -> Optional[str]:
        """
        Name of Athena workgroup.

        Workgroup can set resource limits or override output location.
        """

    @property
    def output_location(self) -> Optional[str]:
        """
        Query output location on S3.

        Can be empty if default location is configured for a workgroup.
        """

    def execute(self, sql: str, *, ignore_cache: bool = False) -> QueryResults:
        """
        Execute a query and wait for results.

        This is a blocking method that waits until query finishes.
        Returns :class:`QueryResults`.

        :param sql: SQL query to be executed
        :param ignore_cache: do not load cached results
        :return: query results
        """
        # Do not override this method.
        # Wrappers do not call execute on the wrapped instance,
        # so overrides are likely to have no effect.
        return self.submit(sql, ignore_cache=ignore_cache).get_results()

    @abstractmethod
    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        """
        Submit a query and return.

        This is a non-blocking method that starts a query and returns.
        Returns a :class:`Query` instance for monitoring query execution
        and downloading results later.

        :param sql: an SQL query to be executed
        :param ignore_cache: do not load cached results
        :return: a query instance
        """

    @abstractmethod
    def get_query(self, execution_id: str) -> Query:
        """
        Get a previously submitted query execution.

        Athena stores results in S3 and does not delete them by default.
        This method can get past queries and retrieve their results.

        :param execution_id: an Athena query execution ID.
        :return: a query instance
        """


class Query(metaclass=ABCMeta):
    """
    Query interface

    Provides access to one query execution.

    Instances of this class are returned by :meth:`Athena.submit`
    and :meth:`Athena.get_query` methods.
    """

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: execution_id={self.execution_id!r}>"

    @property
    @abstractmethod
    def execution_id(self) -> str:
        """
        Athena query execution ID.

        Returns a unique ID of this query execution.
        This ID can be used to retrieve the query using
        the :meth:`.Athena.get_query()` method.
        """

    @abstractmethod
    def get_info(self) -> QueryInfo:
        """
        Retrieve information about this query execution.

        Returns a status of this query with other information.
        """

    @abstractmethod
    def get_results(self) -> QueryResults:
        """
        Retrieve results of this query execution.

        Waits until this query execution finishes and downloads results.
        """

    @abstractmethod
    def kill(self) -> None:
        """
        Kill this query execution.
        """

    @abstractmethod
    def join(self) -> None:
        """
        Wait until this query execution finishes.
        """


class AthenaWrapper(Athena):
    """
    Base class for :class:`.Athena` decorators.
    """

    _wrapped: Athena

    def __init__(self, wrapped: Athena) -> None:
        self._wrapped = wrapped

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.wrapped!r}>"

    @property
    def wrapped(self) -> Athena:
        return self._wrapped

    @property
    def database(self) -> Optional[str]:
        return self._wrapped.database

    @property
    def workgroup(self) -> Optional[str]:
        return self._wrapped.workgroup

    @property
    def output_location(self) -> Optional[str]:
        return self._wrapped.output_location

    def submit(self, sql: str, *, ignore_cache: bool = False) -> Query:
        return self._wrapped.submit(sql, ignore_cache=ignore_cache)

    def get_query(self, execution_id: str) -> Query:
        return self._wrapped.get_query(execution_id)


class QueryWrapper(Query):
    """
    Base class for :class:`.Query` decorators.
    """

    _wrapped: Query

    def __init__(self, wrapped: Query) -> None:
        self._wrapped = wrapped

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.wrapped!r}>"

    @property
    def wrapped(self) -> Query:
        return self._wrapped

    @property
    def execution_id(self) -> str:
        return self._wrapped.execution_id

    def get_info(self) -> QueryInfo:
        return self._wrapped.get_info()

    def get_results(self) -> QueryResults:
        return self._wrapped.get_results()

    def kill(self) -> None:
        self._wrapped.kill()

    def join(self) -> None:
        self._wrapped.join()
